{ Contour Plot Routines (Contour_Plot, Local_Contour) are
 COPYRIGHT 1988, Optimal Systems Laboratory, Plainfield, NJ
  All other routines are COPYRIGHT 1988, John A. Farrell and
  David B. Holtkamp, Los Alamos, NM }

UNIT HiGraf;

{ *** enable only one of these below for directory to look for BGI files *** }
{$DEFINE TPDirectory } { tells InitGraph to look in C:\TP directory }
(* {$DEFINE DefaultDirectory} *) { tells InitGraph to look in
                                 current, default directory }

{ *** enable next line if you want to use TPEnhKbd unit and other Turbo
      Professional units from TurboPower Software*** }
{$DEFINE UseTurboProfessional }

{ *** enable only one line below for Initialize_Graph to work properly *** }
{$DEFINE AUTOMODE}      { autodetect mode }
(* {$DEFINE UseATT}  *) { force ATT 400 or Compaq displays }
(* {$DEFINE Use8514} *) { IBM 8514A display; CAUTION: the XOR calls to
   create the long cross hairs will not work on the 8514.
   There is a bug in 8514 BGI driver or in HDILOAD }
(* {$DEFINE UseVGA} *)  { force VGA mode on display }
(* {$DEFINE UseEGA} *)  { force EGA mode on display }
(* {$DEFINE UseCGA} *)  { force CGA mode on display }

INTERFACE

{ *** set up parameters for using the TPEnhKbd Unit for
      cross hair/cursor control *** }
{$IFDEF UseTurboProfessional}
USES  Dos, TPCrt, TPEnhKbd, Graph;
{ NOTE: if you do not have Turbo Professional TPEnhKbd, there are two line
  (each) in Update_CrossHairs and Update_Cursor that should be commented out;
  you lose the motion for <Ctrl>-Up and <Ctrl>-Down arrow.
  ***CAUTION***: since TPEnhKbd takes over INT $9, you should be careful to
  not have anything that also needs INT $9. This includes the TP5 Inte-
  grated Development Environment debugger, (old) SideKick, and Buf128.
  See Turbo Professional manual on TPEnhKbd and the TP5 README file. }
{$ELSE}
USES Dos, Crt, Graph;
{$ENDIF}

CONST
  Max_X_Size = 256; { maximum width of 2D data array }
  Max_Y_Size = 256; { maximum height of 2D data array }
  Max_Contours = 100; { maximum number of contours }

TYPE
  FLOAT = EXTENDED;
  { definition of floating type:  if coprocessor present, setting
    this to single, double or extended yields rapid operation, while
    setting it to real yields slow operation with a coprocessor.
    If no coprocessor is present, most rapid operation is obtained
    by setting this equal to real }
  LINLOG = (lin,log);

  { definitions for 2D data array type and contour array type }
  Column_Type = ARRAY[0..Max_Y_Size-1] OF LONGINT;
  Column_Pointer_Type = ^Column_Type;
  TwoD_Array_Type = ARRAY[0..Max_X_Size-1] OF Column_Pointer_Type;
  TwoD_Array_Pointer_Type = ^TwoD_Array_Type;
  Contour_Array_Type = ARRAY[0..Max_Contours-1] OF FLOAT;

VAR { these are declared global for use everywhere }
  TwoD_Array_Pointer     :TwoD_Array_Pointer_Type; { this is initialized to
                          nil in Unit init section; no mem allocation is
                          done here }
  Contours               :Contour_Array_Type; { this array of values
                           must be ordered from lowest[0] to
                           highest[Max_Contours-1] }

FUNCTION  Log10(X:FLOAT):FLOAT;
FUNCTION  Pow(X,Y:FLOAT):FLOAT;
PROCEDURE Axes(XOrg,YOrg,XTick,YTick: FLOAT;
               XMajor,YMajor: INTEGER;
               GridOn :BOOLEAN);
PROCEDURE Beep;
PROCEDURE Clear_StatusLine;
PROCEDURE Contour_Plot(X_Begin, X_End, X_Step,
                       Y_Begin, Y_End, Y_Step,
                       Num_Contours :INTEGER;
                       Use_Colors :BOOLEAN);
PROCEDURE Color_Contour_Plot(X_Begin, X_End, X_Step,
                             Y_Begin, Y_End, Y_Step,
                             Num_Contours :INTEGER);
PROCEDURE Draw(X,Y: FLOAT);
PROCEDURE Draw_Label(X,Y: FLOAT; Strng: STRING; Font,Direction,Size,
                     XJust,YJust: WORD);
PROCEDURE Fart;
PROCEDURE Frame(XTick,YTick: FLOAT; XMajor,YMajor:INTEGER; GridOn :BOOLEAN);
PROCEDURE Initialize_CrossHairs;
PROCEDURE Initialize_Cursor;
PROCEDURE Initialize_Graph;
PROCEDURE Label_Axes(LabelX,LabelY: FLOAT);
PROCEDURE Move(X,Y: FLOAT);
PROCEDURE PlotData(X,SigX,Y,SigY: FLOAT; Symbol,SymSize: INTEGER);
PROCEDURE PlotPoint(X,Y: FLOAT; Symbol,SymSize: INTEGER);
PROCEDURE ReadKeyStroke(VAR Keystroke :CHAR; VAR Exkey :BOOLEAN);
PROCEDURE RelDraw(X,Y: FLOAT);
PROCEDURE RelMove(X,Y: FLOAT);
PROCEDURE Remove_Blanks(VAR Strng: STRING);
PROCEDURE Remove_Zeros(VAR Strng: STRING);
PROCEDURE SetClipOn;
PROCEDURE SetClipOff;
PROCEDURE Setup_Graph(XMin,XMax,YMin,YMax,
                      WXMin,WXMax,WYMin,WYMax: FLOAT;
                      logx,logy: LINLOG);
PROCEDURE Toggle_CrossHairs(XScreen,YScreen :INTEGER);
PROCEDURE Toggle_Cursor(XScreen,YScreen :INTEGER);
PROCEDURE Toptitle(Labl: STRING);
PROCEDURE Update_CrossHairs(VAR XScreen,YScreen :INTEGER;
                            VAR Keystroke :CHAR);
PROCEDURE Update_Cursor(VAR XScreen,YScreen :INTEGER;
                        VAR Keystroke :CHAR);
PROCEDURE Window_to_World(XScreen,YScreen:INTEGER;
                          VAR X,Y: FLOAT);
PROCEDURE World_to_Window(X,Y: FLOAT;
                          VAR XScreen,YScreen: INTEGER);
PROCEDURE Write_StatusLine(Message :STRING);
PROCEDURE XAxisTitle(Labl: STRING);
PROCEDURE YAxisTitle(Labl: STRING);

{ ******************************************************************* }

IMPLEMENTATION

CONST
  Custom_LineStyle : ARRAY[0..4] OF WORD =
        ($FFFF,$5555,$0F0F,$1111,$1010);
{
 WORD: 65535: $FFFF:  1111111111111111 (solid line)
 WORD: 21845: $5555:  0101010101010101 (very fine dotted line)
 WORD:  3855: $0F0F:  0000111100001111 (dashed line)
 WORD:  4369: $1111:  0001000100010001 (fine spaced dotted line)
 WORD:  ????: $1010:  0001000100010001 (very finely spaced dotted line)
}

VAR { these are all variables internal to the HiGraf unit }
  XWorldMin, XWorldMax,
  YWorldMin, YWorldMax: FLOAT; { global variables for world limits }
  XWindMin, XWindMax,
  YWindMin, YWindMax: INTEGER; { global variables for window limits }
  XViewMax,YViewMax: INTEGER; { viewport limits }
  XAxisScale, YAxisScale: FLOAT; { factors for converting }
  ConstX, ConstY: LONGINT;       { to screen coordinates }
  XAxisLog, YAxisLog: BOOLEAN; { global variables for axis type }
  MaxLabLength: INTEGER; { y axis label length in pixels; used
                           to calculate y axis title offset }
  MaxLabHeight: INTEGER; { x axis label height in pixels; used
                           to calculate x axis title offset }
  MaxColor: INTEGER; { store GetMaxColor value }
  CrossHairs_Onflag: BOOLEAN; { cross hairs on flag (=TRUE) }
  Cursor_Pointer: POINTER; { pointer for cursor image }
  Cursor_Xoffset,          { X and Y pixel offsets for }
  Cursor_Yoffset: INTEGER; { center of cursor }
  Cursor_Onflag: BOOLEAN;  { cursor on flag (=TRUE) }
  HiGraf_ExitSave: POINTER; { for saving old exit proc address }

{ ******************************************************************* }

PROCEDURE Initialize_Graph;
{
  initialize graphics hardware; do this only once.
}
VAR
  graphdriver,graphmode,errorcode: INTEGER;
BEGIN

{$IFDEF Use8514} { use IBM 8514A for driver }
  GraphDriver := IBM8514;
  GraphMode := IBM8514Hi;
{$ENDIF}

{$IFDEF UseATT} { force ATT400 hires mode }
  graphdriver := ATT400;
  graphmode := ATT400hi;
{$ENDIF}

{$IFDEF UseVGA} { force VGA mode }
  graphdriver := VGA;
  graphmode := VGAHi;
{$ENDIF}

{$IFDEF UseEGA} { force EGA mode }
  graphdriver := EGA;
  graphmode := EGAHi;
{$ENDIF}

{$IFDEF UseCGA} { force CGA mode }
  graphdriver := CGA;
  graphmode := CGAHi;
{$ENDIF}

{$IFDEF AUTOMODE} { enable autodetection of graphics hardware }
  graphdriver := detect;
{$ENDIF}

{$IFDEF TPDirectory}
  InitGraph(graphdriver,graphmode,'C:\TP');
{$ENDIF}

{$IFDEF DefaultDirectory}
  InitGraph(graphdriver,graphmode,'');
{$ENDIF}

  errorcode := graphresult;
  IF errorcode <> GrOK THEN
  BEGIN
    WRITELN('Graphics error: ',GraphErrorMsg(errorcode));
    WRITELN('Program aborted');
    HALT(1);
  END;
END; { Initialize_Graph }

{ ******************************************************************* }

FUNCTION Pow(X,Y:FLOAT):FLOAT;
{
  FUNCTION to compute value of x**y(i.e.
  x raised to the y power
  usage:
        result = Pow(x,y)
        where result, x, and y are FLOAT args
}
BEGIN
  Pow := EXP(Y*LN(X));
END;

{ ******************************************************************* }

FUNCTION Log10(X:FLOAT):FLOAT;
{
  function to take log(base 10) of arg x
  usage:
        result = Log10(x)
        where result and x are FLOAT args
}
CONST
  one_over_ln10 = 0.434294481903252;
BEGIN
  Log10 := LN(X)*one_over_ln10;
END;

{ ******************************************************************* }

PROCEDURE Beep;
{
 short beep to politely alert user
}
BEGIN
  Sound(1000);
  Delay(100);
  NoSound;
END;

{ ******************************************************************* }

PROCEDURE Fart;
{
 a less than polite way of bringing something to the users attention
}
BEGIN
   Sound(50);
   Delay(1000);
   NoSound;
END;

{ ******************************************************************* }

PROCEDURE SetClipOn;
{
  sets clipping of plotting on: i.e. only within actual plot region
  delimited by XMin->XMax and YMin->YMax
}
BEGIN
  SetViewPort(XWindMin,YWindMin,XWindMax,YWindMax,ClipOn)
END; { SetClipOn }

{ ******************************************************************* }

PROCEDURE SetClipOff;
{
  sets clipping of plotting off: i.e. can plot anywhere on the screen,
  but window is still delimited by XMin->XMax and YMin->YMax
}
BEGIN
  SetViewPort(XWindMin,YWindMin,XWindMax,YWindMax,ClipOff)
END; { SetClipOff }

{ ******************************************************************* }

PROCEDURE Clear_StatusLine;
{
  clears a status line at the very top of the screen;
  you have to be in graphics mode for this to work.
}
CONST
  Status_Width = 12; { is status line 12 pixels tall? }
BEGIN
  SetViewPort(0,0,GetMaxX,GetMaxY,ClipOff);
  SetFillStyle(EmptyFill,GetMaxColor);
  Bar(0,0,GetMaxX,Status_Width);
  SetClipOn;
END; { of Clear_StatusLine }

{ ******************************************************************* }

PROCEDURE Write_StatusLine(Message :STRING);
{
  writes out a status line "Message" after clearing the status line
  you have to be in graphics mode for this to work.
}
BEGIN
  Clear_StatusLine;
  SetViewPort(0,0,GetMaxX,GetMaxY,ClipOff);
  SetTextStyle(SmallFont,HorizDir,4);
  SetTextJustify(LeftText,TopText);
  MoveTo(0,0);
  OutText(Message);
  SetClipOn;
  GoToXY(1,2);
END;

{ ******************************************************************* }

PROCEDURE Setscale(WXMin,WXMax,WYMin,WYMax: FLOAT);
{
  defines constants (XAxisScale,YAxisScale,ConstX,ConstY) which are used
  to convert to and from world and window coordinate systems; this is
  used only internally to HiGraf.
 }
BEGIN
  XWindMin := ROUND(WXMin*GetMaxX/100.0);
  XWindMax := ROUND(WXMax*GetMaxX/100.0);
  YWindMin := ROUND(WYMin*GetMaxY/100.0);
  YWindMax := ROUND(WYMax*GetMaxY/100.0);
  XViewMax := XWindMax-XWindMin;
  YViewMax := YWindMax-YWindMin;
  IF (XAxisLog) THEN
  BEGIN
    XAxisScale := XViewMax/(Log10(XWorldMax)-Log10(XWorldMin));
    ConstX := -ROUND(Log10(XWorldMin)*XAxisScale);
  END
  ELSE
  BEGIN
    XAxisScale := XViewMax/(XWorldMax-XWorldMin);
    ConstX := -ROUND(XWorldMin*XAxisScale);
  END; { ELSE }
  IF (YAxisLog) THEN
  BEGIN
    YAxisScale := YViewMax/(Log10(YWorldMin)-Log10(YWorldMax));
    ConstY := -ROUND(Log10(YWorldMax)*YAxisScale);
  END
  ELSE
  BEGIN
    YAxisScale := YViewMax/(YWorldMin-YWorldMax);
    ConstY := -ROUND(YWorldMax*YAxisScale);
  END; { ELSE }
END; { Setscale }

{ ******************************************************************* }

PROCEDURE World_to_Window(X,Y: FLOAT; VAR XScreen,YScreen: INTEGER);
{
  performs transformation from world coordinates to window (screen)
  coordinates
}
VAR
  temp :FLOAT;
  isign :INTEGER;
BEGIN
  IF (XAxisLog) THEN
  BEGIN
    IF (x > 0.0) THEN x := Log10(x)
    ELSE x := Log10(XWorldMin);
  END;
  IF (YAxisLog) THEN
  BEGIN
    IF (y > 0.0) THEN y := Log10(y)
    ELSE y := Log10(YWorldMin);
  END;

  temp := x*XAxisScale + ConstX;
  IF (temp > 0) THEN isign := 1
    ELSE isign := -1;
  IF (isign*temp < MaxInt) THEN
    XScreen := ROUND(temp)
  ELSE XScreen := isign*MaxInt;

  temp := y*YAxisScale + ConstY;
  IF (temp > 0) THEN isign := 1
    ELSE isign := -1;
  IF (isign*temp < MaxInt) THEN
    YScreen := ROUND(temp)
  ELSE YScreen := isign*MaxInt;

END; { World_to_Window }

{ ******************************************************************* }

PROCEDURE Window_to_World(XScreen,YScreen: INTEGER; VAR X,Y: FLOAT);
{
  performs transformation from window (screen) coordinates to world
  coordinates
}
BEGIN
  X := (1.0*XScreen-ConstX)/XAxisScale;
  Y := (1.0*YScreen-ConstY)/YAxisScale;
  IF (XAxisLog) THEN X := Pow(10.0,x);
  IF (YAxisLog) THEN Y := Pow(10.0,y);
END; { Window_to_World }

{ ******************************************************************* }

PROCEDURE Move(X,Y: FLOAT);
{
  move graphics cursor to world coordinate X,Y
}
VAR xs,ys: INTEGER;
BEGIN
  World_to_Window(X,Y,xs,ys);
  MoveTo(xs,ys)
END; { Move }

{ ******************************************************************* }

PROCEDURE Draw(X,Y: FLOAT);
{
  draw line from present position of graphics cursor to world
  coordinate X,Y; use BGI procedure SetLineStyle to change from
  solid to dotted, dashed, etc
}
VAR xs,ys: INTEGER;
BEGIN
  World_to_Window(X,Y,xs,ys);
  LineTo(xs,ys)
END; { Draw }

{ ******************************************************************* }

PROCEDURE RelMove(X,Y: FLOAT);
{
  move graphics cursor a relative distance (in world coordinates) from
  present position
}
VAR xs,ys: INTEGER;
BEGIN
  World_to_Window(X,Y,xs,ys);
  MoveRel(xs,ys)
END; { RelMove }

{ ******************************************************************* }

PROCEDURE RelDraw(X,Y: FLOAT);
{
  draw line from present position of graphics cursor to a new
  position given by the relative distance X,Y (in world coordinates);
  use BGI procedure SetLineStyle to change from solid to dotted,
  dashed, etc
}
VAR xs,ys: INTEGER;
BEGIN
  World_to_Window(X,Y,xs,ys);
  LineRel(xs,ys)
END; { RelDraw }

{ ****************************************************************** }

PROCEDURE Tick_Proc(Xarg,Yarg,XTick,YTick: FLOAT;
               XMajor,YMajor: INTEGER;
               GridOn :BOOLEAN);
{
  Draws tick marks for "axes" and "frame" with minor tick intervals
  every XTick,YTick, and major ticks every XMajor,YMajor minor ticks.
  Tick_Proc checks if x and y axes are running the conventional directions
  (i.e. xmin->max is left to right and ymin->ymax is up) and acts accordingly;
  This routine is only used internally by HiGraf.
}
CONST
  XTick_scale = 0.015;     { scale factors for tick lengths }
  YTick_scale = 0.005;
VAR
  x, y, decade, tempmin, tempmax: FLOAT;
  ticklen, i: INTEGER;
BEGIN
{ draw x axis tick marks }
  ticklen := ROUND(XTick_scale*GetMaxY);
  IF (XWorldMin <= XWorldMax) THEN { normal axis direction }
  BEGIN
    tempmin := XWorldMin;
    tempmax := XWorldMax;
  END
  ELSE { axis is reversed from normal }
  BEGIN
    tempmin := XWorldMax;
    tempmax := XWorldMin;
  END; { IF (XWorldMin... }

  IF (XAxisLog) THEN { x axis is logarithmic }
  BEGIN
    decade := Pow(10.0,TRUNC(Log10(tempmin))); { get next lower decade }
    REPEAT
      Move(decade,Yarg);
      LineRel(0,2*ticklen);
      LineRel(0,-4*ticklen);
      IF (GridOn) THEN
      BEGIN
        Move(decade,YWorldMin);
        Draw(decade,YWorldMax);
      END; { IF GridOn }
      FOR i :=  2 TO 9 DO
      BEGIN
        x := decade*i;
        IF (x<= tempmax) THEN
        BEGIN
          Move(x,Yarg);
          LineRel(0,ticklen);
          LineRel(0,-2*ticklen);
        END;
        IF (GridOn) THEN
        BEGIN
          SetLineStyle(UserBitLn,Custom_LineStyle[3],1); { fine dotted line }
          Move(x,YWorldMin);
          Draw(x,YWorldMax);
          SetLineStyle(SolidLn,0,NormWidth);
        END; { IF (GridOn) }
      END; { FOR i }
      decade := decade*10.0;
    UNTIL (decade > tempmax);
  END { IF (XAxisLog)... }

  ELSE IF (XTick > 0.0) THEN { x axis is linear }
  BEGIN
    x := tempmin+XTick;
    i := 1;
    REPEAT
      Move(x,Yarg);
      IF (i MOD XMajor = 0) THEN
      BEGIN
        LineRel(0,2*ticklen);
        LineRel(0,-4*ticklen);
      END
      ELSE BEGIN
        LineRel(0,ticklen);
        LineRel(0,-2*ticklen);
      END;
      IF (GridOn) THEN
      BEGIN
        IF (i MOD XMajor <> 0) THEN
          SetLineStyle(UserBitLn,Custom_LineStyle[3],1); { fine dotted line }
        Move(x,YWorldMin);
        Draw(x,YWorldMax);
        SetLineStyle(SolidLn,0,NormWidth);
      END;
      x := x+XTick;
      INC(i);
    UNTIL (x > tempmax);
  END; { ELSE IF (XTick... }

  { draw y axis tick marks }
  ticklen := ROUND(YTick_scale*GetMaxX);
  IF (YWorldMin <= YWorldMax) THEN { normal axis direction }
  BEGIN
    tempmin := YWorldMin;
    tempmax := YWorldMax;
  END
  ELSE { axis is reversed from normal }
  BEGIN
    tempmin := YWorldMax;
    tempmax := YWorldMin;
  END; { IF (YWorldMin... }

  IF (YAxisLog) THEN { y axis is logarithmic }
  BEGIN
    decade := Pow(10.0,TRUNC(Log10(tempmin))); { get next lowest decade }
    REPEAT
      Move(Xarg,decade);
      LineRel(-2*ticklen,0);
      LineRel(4*ticklen,0);
      IF (GridOn) THEN
      BEGIN
        Move(XWorldMin,decade);
        Draw(XWorldMax,decade);
      END; { IF (GridOn) }
      FOR i := 2 TO 9 DO
      BEGIN
        y := decade*i;
        IF (y <= tempmax) THEN
        BEGIN
          Move(Xarg,y);
          LineRel(-ticklen,0);
          LineRel(2*ticklen,0);
        END;
        IF (GridOn) THEN
        BEGIN
          SetLineStyle(UserBitLn,Custom_LineStyle[3],1); { fine dotted line }
          Move(XWorldMin,y);
          Draw(XWorldMax,y);
          SetLineStyle(SolidLn,0,NormWidth);
        END; { IF (GridOn) }
      END; { FOR i }
      decade := decade*10.0;
    UNTIL (decade > tempmax);
  END { IF (YAxisLog) }

  ELSE IF (YTick > 0.0) THEN { y axis is linear }
  BEGIN
    y := tempmin+YTick;
    i := 1;
    REPEAT
      Move(Xarg,y);
      IF (i MOD YMajor = 0) THEN
      BEGIN
        LineRel(-2*ticklen,0);
        LineRel(4*ticklen,0);
      END
      ELSE
      BEGIN
        LineRel(-ticklen,0);
        LineRel(2*ticklen,0);
      END;
      IF (GridOn) THEN
      BEGIN
        IF (i MOD YMajor <> 0) THEN
          SetLineStyle(UserBitLn,Custom_LineStyle[3],1); { fine dotted line }
        Move(XWorldMin,y);
        Draw(XWorldMax,y);
        SetLineStyle(SolidLn,0,NormWidth);
      END;
      y := y+YTick;
      INC(i);
    UNTIL y > tempmax;
  END; { ELSE IF (YTick... }

END; { Tick_Proc }

{ ******************************************************************* }

PROCEDURE Frame(XTick,YTick: FLOAT;
                XMajor,YMajor: INTEGER;
                GridOn :BOOLEAN);
{
  draws a frame around box determined by XMin,XMax and YMin,YMax;
  put minor ticks every XTick,YTick interval with major ticks every
  XMajor,YMajor multiples of XTick,YTick; if GridOn = TRUE a grid
  is drawn
}
BEGIN
  SetClipOff;

  { draw frame around world }
  Move(XWorldMin,YWorldMin);
  Draw(XWorldMax,YWorldMin);
  Draw(XWorldMax,YWorldMax);
  Draw(XWorldMin,YWorldMax);
  Draw(XWorldMin,YWorldMin);
  SetClipOn;

  { add tick marks }
  Tick_Proc(XWorldMin,YWorldMin,XTick,YTick,XMajor,YMajor,GridOn);
  Tick_Proc(XWorldMax,YWorldMax,XTick,YTick,XMajor,YMajor,GridOn);
END; { Frame }

{ ******************************************************************* }

PROCEDURE Remove_Blanks(VAR Strng: STRING);
{
  This PROCEDURE removes the leading blanks from a string.
}
CONST remove: SET OF CHAR = [#8,#9,#10,#12,#13,' '];
BEGIN
  WHILE (Strng[1] IN remove) AND (LENGTH(Strng) > 0) DO DELETE(Strng,1,1);
END; { Remove_Blanks }

{ ****************************************************************** }

PROCEDURE Remove_Zeros(VAR Strng: STRING);
{
  This PROCEDURE removes trailing zeros, blanks, and the decimal
  point (if followed by all zeros) from a string. It will, in its
  own kludgy way, adjust labels like "1.0E+0101" to be "1E+101"
  and the like. This routine is not elegant, but it does seem to
  work in all cases (at least the ones I've come across...)
}
CONST
  remove: SET OF CHAR = [#8,#9,#10,#12,#13,' ','0'];
  numbers: SET OF CHAR = ['1','2','3','4','5','6','7','8','9'];
VAR
  i,j,k: INTEGER;
BEGIN
  i := LENGTH(Strng);
  j := i; { store old length of string }
  WHILE (Strng[i] IN remove) AND (i > 0) DO   { delete trailing zeros }
  BEGIN
    DELETE(Strng,i,1);
    DEC(i);
  END;
  { now make sure I have not clobbered something like 1.0E+0100 into 1.E+01;
    but check to make sure there is an 'E' in there somewhere }
  k := j-i;
  IF ((k > 0) AND (Strng[i] IN numbers) AND (POS('E',Strng) > 0)) THEN
  { yep it happened; put it back }
    CASE (k) OF
     1: Strng := Strng + '0';
     2: Strng := Strng + '00';
     3: Strng := Strng + '000';
    END; { of CASE }
  i := LENGTH(Strng); { adjust string length }

  IF (Strng[i] = '.') THEN { last character is a "." }
  BEGIN
    DELETE(Strng,i,1);    { delete decimal point }
    DEC(i);
  END;

  { the problem with this one is I can't know if the others have 4 digit
    exponents; so assume you do not have a 100 decade log plot so its ok }
  IF (Strng[i] = '+') THEN { had removed all zeroes from end of exponent }
  BEGIN
    Strng := Strng+'00'; { fix it by adding zeroes back in }
    INC(i,2); { adjust counter }
  END;

  IF (i > 8) THEN { got a long exponent label i.e. 1.0E+0100 or 1.0E+0020 }
  BEGIN { to decide which it is }
    IF ((Strng[i-2] = '0') AND (Strng[i-3] = '0')) THEN { it is a 1.0E+0020 }
    BEGIN
      DELETE(Strng,i-3,2); { rid of extra zeroes in exponent }
      DEC(i,2); { adjust counter }
    END;
  END;
  IF (i > 6) THEN
    IF (Strng[i-3] = 'E') THEN DELETE(Strng,i-5,2) { delete extra ".0" }
    ELSE IF (Strng[i-5] = 'E') THEN DELETE(Strng,i-7,2); { delete extra ".0" }

  i := LENGTH(Strng); { adjust length again }

  IF (i > 6) THEN { check to see if it looks like 1E+0101 }
    IF ((Strng[i-3] = '0') AND (Strng[i-5] = 'E')) THEN
    { this guy looks like 1E+0101 }
    BEGIN
      DELETE(Strng,i-3,1); { make it like 1E+101 }
      i := LENGTH(Strng);
    END;

  IF ((i=0) OR (Strng = '-0')) THEN { by this point,
  if there is nothing left, I really want a 0 in the label! <sigh> }
    Strng := '0';
END; { Remove_Zeros }

{ ****************************************************************** }

PROCEDURE TopTitle(Labl: STRING);
{
  puts STRING Labl at the top of the plot
}
CONST scale = 6;
VAR xw,yw: INTEGER;
BEGIN
  SetClipOff;
  xw := XViewMax DIV 2;
  yw := -scale;
  SetTextStyle(TriplexFont,HorizDir,1);
  SetTextJustify(CenterText,BottomText);
  OutTextXY(xw,yw,Labl);
  SetClipOn;
END; { TopTitle }

{ ****************************************************************** }

PROCEDURE XAxisTitle(Labl: STRING);
{
  labels x axis with STRING Labl; Label_Axes sets up variable MaxLabHeight
  which is used for XAxisTitle offset; if you haven't called Label_Axes,
  Setup_Graph has initialized MaxLabHeight to 0;
}
CONST scale = 10;
VAR xw,yw: INTEGER;
BEGIN
  SetClipOff;
  xw := XViewMax DIV 2;
  yw := YViewMax+MaxLabHeight+scale;
  SetTextStyle(TriplexFont,HorizDir,1);
  SetTextJustify(CenterText,TopText);
  OutTextXY(xw,yw,Labl);
  SetClipOn;
END; { XAxisTitle }

{ ****************************************************************** }

PROCEDURE YAxisTitle(Labl: STRING);
{
  labels y axis with STRING Labl; Label_Axes sets up variable MaxLabLength
  which is used for YAxisTitle offset; if you haven't called Label_Axes,
  Setup_Graph has initialized MaxLabLength to 0;
}
CONST scale = 20;
VAR xw,yw: INTEGER;
BEGIN
  SetClipOff;
  xw := -MaxLabLength-scale; { adjust position for length of y axis number labels }
  yw := YViewMax DIV 2;
  SetTextStyle(TriplexFont,VertDir,1);
  SetTextJustify(RightText,CenterText);
  OutTextXY(xw,yw,Labl);
  SetClipOn;
END; { YAxisTitle }

{ ****************************************************************** }

PROCEDURE Draw_Label(X,Y: FLOAT; Strng: STRING; Font,Direction,Size,
                     XJust,YJust: WORD);
{
  This PROCEDURE draws a label at position X,Y in world coordinates;
  the variables "Font,Direction,Size,XJust,YJust" are just as in the
  BGI description for Turbo Pascal graphics:
  Font,Direction, and Size are set by a BGI proc "SetTextStyle" and
  XJust,YJust are set by a BGI proc "SetTextJustify"
}
VAR xw,yw: INTEGER;
BEGIN
  SetTextJustify(XJust,YJust);
  SetTextStyle(Font,Direction,Size);
  World_to_Window(X,Y,xw,yw);
  OutTextXY(xw,yw,Strng);
END; { Draw_Label }

{ ****************************************************************** }

PROCEDURE Axes(XOrg,YOrg,XTick,YTick: FLOAT;
               XMajor,YMajor: INTEGER;
               GridOn :BOOLEAN);
{
  Draws a pair of axes through XOrg,YOrg with minor tick intervals
  XTick,YTick, and major ticks every XMajor,YMajor minor tick
  intervals; if the flag GridOn = TRUE then a grid is drawn
}
BEGIN
  SetClipOn;
  Move(XOrg,YWorldMin); Draw(XOrg,YWorldMax);  { x axis }
  Move(XWorldMin,YOrg); Draw(XWorldMax,YOrg);  { y axis }
  Tick_Proc(XOrg,YOrg,XTick,YTick,XMajor,YMajor,GridOn); { draw tick marks }
END; { Axes }

{ ****************************************************************** }

PROCEDURE Label_Axes(LabelX,LabelY: FLOAT);
{
  Labels axes at intervals LabelX, LabelY; by convention you may
  want this to be :
      LabelX = XMajor*XTick
      LabelY = YMajor*YTick
  in your program
}
CONST
  x_location = 8;  { factors for positioning labels }
  y_location = 8;
VAR
  xloc,yloc,templength,tempheight: INTEGER;
  tempminx,tempmaxx,tempminy,tempmaxy: FLOAT;
  x,y: FLOAT;
  labl: STRING;
BEGIN
  SetClipOff;
  SetTextStyle(SmallFont,HorizDir,4);

  IF (XWorldMin <= XWorldMax) THEN { regular axis direction }
  BEGIN
    IF (XAxisLog) THEN
    BEGIN { adjust limits to ensure end labels are shown }
      World_to_Window(XWorldMin,YWorldMin,xloc,yloc);
      Window_to_World(xloc-1,yloc,tempminx,y);
      World_to_Window(XWorldMax,YWorldMin,xloc,yloc);
      Window_to_World(xloc+1,yloc,tempmaxx,y);
    END
    ELSE { X axis is linear }
    BEGIN
      tempminx := XWorldMin;
      tempmaxx := XWorldMax;
    END;
  END
  ELSE { reversed axis direction }
  BEGIN
    IF (XAxisLog) THEN
    BEGIN { adjust limits to ensure end labels are shown }
      World_to_Window(XWorldMax,YWorldMin,xloc,yloc);
      Window_to_World(xloc+1,yloc,tempminx,y);
      World_to_Window(XWorldMin,YWorldMin,xloc,yloc);
      Window_to_World(xloc-1,yloc,tempmaxx,y);
    END
    ELSE { X axis is linear }
    BEGIN
      tempminx := XWorldMax;
      tempmaxx := XWorldMin;
    END;
  END;

  IF (YWorldMin <= YWorldMax) THEN { regular axis direction }
  BEGIN
    IF (YAxisLog) THEN
    BEGIN { adjust limits to ensure end labels are shown }
      World_to_Window(XWorldMin,YWorldMin,xloc,yloc);
      Window_to_World(xloc,yloc+1,x,tempminy);
      World_to_Window(XWorldmin,YWorldMax,xloc,yloc);
      Window_to_World(xloc,yloc-1,x,tempmaxy);
    END
    ELSE { Y axis is linear }
    BEGIN
     tempminy := YWorldMin;
     tempmaxy := YWorldMax;
   END;
  END
  ELSE { reversed axis direction }
  BEGIN
    IF (YAxisLog) THEN
    BEGIN { adjust limits to ensure end labels are shown }
      World_to_Window(XWorldMin,YWorldMax,xloc,yloc);
      Window_to_World(xloc,yloc-1,x,tempminy);
      World_to_Window(XWorldMin,YWorldMin,xloc,yloc);
      Window_to_World(xloc,yloc+1,x,tempmaxy);
    END
    ELSE { Y axis is linear }
    BEGIN
      tempminy := YWorldMax;
      tempmaxy := YWorldMin;
    END;
  END;

  { label x axis }
  MaxLabHeight := 0; { initialize this offset for XAxisTitle }
  tempheight := 0;
  SetTextJustify(CenterText,TopText); { do this just once for linear axis }
  x := tempminx;
  IF (XAxisLog) THEN
    x := Pow(10,TRUNC(Log10(tempminx)));

  WHILE (x <= tempmaxx) DO
  BEGIN
    IF (XAxisLog) THEN STR(TRUNC(Log10(x)):6,labl)
    ELSE STR(x:10:5,labl);
    Remove_Blanks(labl);
    Remove_Zeros(labl);
    World_to_Window(x,YWorldMin,xloc,yloc);
    IF (XAxisLog) THEN { do an exponentiated label }
    BEGIN
      IF ((x >= tempminx) AND (x <= tempmaxx)) THEN
      BEGIN
        SetTextJustify(CenterText,TopText);
        yloc := yloc+y_location; { position the '10' }
        OutTextXY(xloc,yloc,'10');
        tempheight := TextHeight('10');
        yloc := yloc-(tempheight DIV 2);
        xloc := xloc+(TextWidth('10') DIV 2)+3;
        tempheight := (tempheight DIV 2) + TextHeight(labl); { pixel height of label }
        SetTextJustify(LeftText,TopText);
        OutTextXY(xloc,yloc,labl);
      END; { IF ((x >= tempminx... }
      x := x*10;
    END
    ELSE { straight linear label }
    BEGIN
      yloc := yloc+y_location;
      OutTextXY(xloc,yloc,labl);
      tempheight := TextHeight(labl);
      x := x+LabelX;
    END; { IF (XAxisLog)... }
    IF (tempheight > MaxLabHeight) THEN MaxLabHeight := tempheight;
  END; { WHILE }

  { label y axis }
  MaxLabLength := 0; { initialize this offset for YAxisTitle }
  templength := 0;
  SetTextJustify(RightText,CenterText);
  y := tempminy;
  IF (YAxisLog) THEN
    y := Pow(10,TRUNC(Log10(tempminy)));
  WHILE (y <= tempmaxy) DO
  BEGIN
    IF (YAxisLog) THEN STR(TRUNC(log10(y)):6,labl)
    ELSE STR(y:10:5,labl);
    Remove_Blanks(labl);
    Remove_Zeros(labl);
    World_to_Window(XWorldMin,y,xloc,yloc);
    IF (YAxisLog) THEN { do an exponentiated label }
    BEGIN
      IF ((y >= tempminy) AND (y <= tempmaxy)) THEN
      BEGIN
        SetTextJustify(RightText,CenterText);
        templength := TextWidth('10');
        xloc := xloc-x_location-templength; { position the '10' }
        OutTextXY(xloc,yloc,'10');
        templength := templength+TextWidth(labl); { pixel width of label }
        yloc := yloc-(TextHeight('10') DIV 2);
        xloc := xloc+2;
        SetTextJustify(LeftText,CenterText);
        OutTextXY(xloc,yloc,labl);
      END; { IF ((y >= tempminy... }
      y := y*10;
    END
    ELSE { straight linear label }
    BEGIN
      xloc := xloc-x_location;
      OutTextXY(xloc,yloc,labl);
      templength := TextWidth(labl);
      y := y+LabelY;
    END;
    IF (templength > MaxLabLength) THEN MaxLabLength := templength;
  END; { WHILE }

  { restore defaults }
  SetClipOn;
END; { Label_Axes }

{ ****************************************************************** }

PROCEDURE PlotPoint(X,Y: FLOAT; Symbol,SymSize: INTEGER);
{
  Plots a Symbol, scaled by SymSize, centered at world coordinates
  X,Y; available Symbols are:
    1  - plus
    2  - box
    3  - cross (x-like)
    4  - triangle
    5  - inverted triangle
    6  - open circle
    7  - filled circle
    8  - dot
}
CONST del = 1;
VAR xs,ys,delta: INTEGER;
BEGIN
  World_to_Window(X,Y,xs,ys);
  delta := del*SymSize;
  CASE Symbol OF
    1: BEGIN  { plus }
         MoveTo(xs-delta,ys);
         LineRel(2*delta,0);
         MoveRel(-delta,-delta);
         LineRel(0,2*delta);
       END;
    2: BEGIN  { box }
         MoveTo(xs-delta,ys-delta);
         LineRel(2*delta,0);
         LineRel(0,2*delta);
         LineRel(-2*delta,0);
         LineRel(0,-2*delta);
       END;
    3: BEGIN  { cross }
         MoveTo(xs-delta,ys-delta);
         LineRel(2*delta,2*delta);
         MoveRel(-2*delta,0);
         LineRel(2*delta,-2*delta);
       END;
    4: BEGIN  { triangle }
         MoveTo(xs,ys-delta-delta DIV 3);
         LineRel(-delta,2*delta);
         LineRel(2*delta,0);
         LineRel(-delta,-2*delta);
       END;
    5: BEGIN  { inverted triangle }
         MoveTo(xs,ys+delta+delta DIV 3);
         LineRel(-delta,-2*delta);
         LineRel(2*delta,0);
         LineRel(-delta,2*delta);
       END;
    6: Circle(xs,ys,delta); { open circle }
    7: FillEllipse(xs,ys,delta,delta); { closed circle }
    8: PutPixel(xs,ys,MaxColors); { unscaled point }
  END; { CASE }
END;  { PlotPoint }

{ ****************************************************************** }

PROCEDURE Setup_Graph(XMin,XMax,YMin,YMax,
                      WXMin,WXMax,WYMin,WYMax: FLOAT;
                      logx,logy: LINLOG);
{
  Procedure to set up the graph limits.
  The world coordinate limits are XMin, XMax, YMin, and YMax.
  The window limits are WXMin, WXMax, WYMin, WYMax expressed as
  fractions (of 1,i.e. 0.95 etc) of the maximum screen limits.
}
BEGIN
  IF (logx = log)
    THEN XAxisLog := TRUE
    ELSE XAxisLog := FALSE;
  IF (logy = log)
    THEN YAxisLog := TRUE
    ELSE YAxisLog := FALSE;

  { set global variables for other graphics routines. }
  XWorldMin := XMin;
  XWorldMax := XMax;
  YWorldMin := YMin;
  YWorldMax := YMax;
  Setscale(WXMin,WXMax,WYMin,WYMax);

  { initialize some global constants just in case Label_Axes is not called }
  MaxLabLength := 0;
  MaxLabHeight := 0;

  { set up defaults for graphics parameters }
  SetClipOn;
END; { Setup_Graph }

{ ****************************************************************** }

PROCEDURE PlotData(X,SigX,Y,SigY: FLOAT; Symbol,SymSize: INTEGER);
{
  Plots a Symbol, scaled by SymSize, centered at world coordinates
  X,Y; error bar on X,Y are sigx and sigy; available Symbols are:
    1  - plus
    2  - box
    3  - cross (x-like)
    4  - triangle
    5  - inverted triangle
    6  - open circle
    7  - filled circle
    8  - dot
}
CONST del = 1;
VAR xs,ys,delta: INTEGER;
BEGIN
  PlotPoint(X,Y,Symbol,SymSize);
  delta := del*SymSize;

  IF (sigx > 0.0) THEN { do x error bar }
  BEGIN
    Move(x+sigx,y);
    Draw(x-sigx,y);
    World_to_Window(x+sigx,y,xs,ys);
    MoveTo(xs,ys-delta);
    LineRel(0,2*delta);
    World_to_Window(x-sigx,y,xs,ys);
    MoveTo(xs,ys-delta);
    LineRel(0,2*delta);
    END; { IF sigx > 0 }

  IF (sigy > 0.0) THEN { do y error bar }
  BEGIN
    Move(x,y+sigy);
    Draw(x,y-sigy);
    World_to_Window(x,y+sigy,xs,ys);
    MoveTo(xs-delta,ys);
    LineRel(2*delta,0);
    World_to_Window(x,y-sigy,xs,ys);
    MoveTo(xs-delta,ys);
    LineRel(2*delta,0);
    END; { IF sigy > 0 }
END;  { PlotData }

{ ****************************************************************** }

PROCEDURE Initialize_Cursor;
{
  this proc initializes the cursor image and calculates some offset
  variables which are used in other cursor routines; also sets
  cursor on flag to FALSE which is == off;
}
CONST
  r  = 5; { length of cursor in x direction in pixels }
  StartX = 100; { x coordinate for initial cursor }
  StartY = 50; { y coordinate for initial cursor }
VAR
  xasp,yasp         : WORD; { words to store aspect ratio info }
  ry                : INTEGER; { length of cursor after correcting for aspect ratio }
  ulx, uly          : WORD; { upper left x and y of image }
  lrx, lry          : WORD; { lower right x and y of cursor image }
  Size              : WORD; { Size of image of cursor in bytes }
BEGIN
  { clear off the screen }
  ClearDevice;

  { calculate aspect ratio of pixels }
  GetAspectRatio(xasp,yasp);
  ry := ROUND(r*1.0*yasp/(1.0*xasp));

  { draw cursor }
  Line(StartX, StartY-ry, StartX, StartY+ry);
  Line(StartX-r, StartY, StartX+r, StartY);

  { read cursor image }
  ulx := StartX-(r+1);
  uly := StartY-(ry+1);
  lrx := StartX+(r+1);
  lry := StartY+(ry+1);

  { get the image and assign cursor_pointer to point at it }
  Size := ImageSize(ulx, uly, lrx, lry);
  GetMem(Cursor_Pointer, Size);
  GetImage(ulx, uly, lrx, lry, Cursor_Pointer^);
  PutImage(ulx, uly, Cursor_Pointer^, XORPut); { erase image }

  { calculate x and y offsets in pixels to calculate center from
    upper left corner of cursor image origin }
  Cursor_Xoffset := r + 1;
  Cursor_Yoffset := ry + 1;

  { set cursor status flag to off }
  Cursor_Onflag := FALSE;
END; { of Initialize_Cursor }

{ ****************************************************************** }

PROCEDURE ReadKeyStroke(VAR Keystroke :CHAR; VAR Exkey :BOOLEAN);
{
  this proc reads any keystroke and returns the character if it is
  a normal ASCII character; the flag Exkey is returned TRUE if the key
  happens to be an extended key scan code, then Keystroke returns
  the second character of the extended code sequence
}
BEGIN
  keystroke := ReadKey;
  IF (keystroke <> #0) THEN { it is not an extended scan code }
    Exkey := FALSE { not an extended scan key }
  ELSE
  BEGIN
    keystroke := Readkey; { get the other character }
    Exkey :=TRUE; { set the extended key flag }
  END;
END; { of Read_KeyStroke }

{ ****************************************************************** }

PROCEDURE Toggle_Cursor(XScreen,YScreen :INTEGER);
{
  this proc toggles cursor status as to whether it is visible; if it
  is on, turn it off; if it is off, turn it on; reset flag accordingly;
  do it at the current cursor position (XScreen,YScreen);
  note that the (XScreen,YScreen) pair is the cursor center, not the
  corner;
}
BEGIN
  { alter x and y pixel address for cursor offset; since (XScreen,YScreen)
   are not VAR into this proc, these changes are only in the scope of
    this proc }
  XScreen := XScreen - Cursor_Xoffset;
  YScreen := YScreen - Cursor_Yoffset;

  IF (Cursor_Onflag) THEN
  BEGIN
    Cursor_Onflag := FALSE; { set flag to cursor off state }
    PutImage(XScreen, YScreen, Cursor_Pointer^, XORPut); { erase image }
  END
  ELSE
  BEGIN
    Cursor_Onflag := TRUE; { set flag for cursor on }
    PutImage(XScreen, YScreen, Cursor_Pointer^, XORPut); { draw image }
  END;
END; { of Toggle_Cursor }

{ ****************************************************************** }

PROCEDURE Update_Cursor(VAR XScreen,YScreen :INTEGER;
                        VAR Keystroke :CHAR);
{
  this routine moves and updates the cursor position; it puts updated
  cursor position in XScreen,YScreen; it assumes on entry
  that the cursor is on and visible; it returns with the ASCII character
  struck in Keystroke, unless the key was a select few of the extended
  scan keys: (see below)
    left/right arrow        - move one screen pixel left/right
    up/down arrow           - move one screen pixel up/down
    home/end                - move one pixel to the left *and* one
                              up/down
    pgup/pgdn               - move one pixel to the right *and* one
                              up/down
    <Ctrl> any of the above - moves the self similar direction but
                              by 10 pixels

  Update_Cursor exits with the cursor off (not visible);
}
VAR
  flag,exkey      :BOOLEAN;
BEGIN
  flag:=FALSE; { initialize done flag on entry }

  { if cursor is on, turn it off }
  IF ( Cursor_Onflag) THEN Toggle_Cursor(XScreen,YScreen); { this XScreen
      and YScreen are not offset }

  { alter XScreen and YScreen pixel address for cursor offset; since
    (XScreen,YScreen) are VAR in this proc, one needs to change them
    back on exit }
  XScreen := XScreen - Cursor_Xoffset;
  YScreen := YScreen - Cursor_Yoffset;

  REPEAT
    PutImage(XScreen, YScreen, Cursor_Pointer^, XORPut); { draw image }
    ReadKeyStroke(keystroke,exkey); { wait for key to be hit }
    PutImage(XScreen, YScreen, Cursor_Pointer^, XORPut); { erase image }
    IF (exkey) THEN { this an extended key scan code }
      CASE keystroke OF
        'H': DEC(YScreen); { up arrow }
        'P': INC(YScreen); { down arrow }
        'K': DEC(XScreen); { left arrow }
        'M': INC(XScreen); { right arrow }
        'G': BEGIN DEC(YScreen); DEC(XScreen); END; { home key }
        'O': BEGIN INC(YScreen); DEC(XScreen); END; { end key }
        'I': BEGIN DEC(YScreen); INC(XScreen); END; { page up }
        'Q': BEGIN INC(YScreen); INC(XScreen); END; { page down }
        's': DEC(XScreen,10); { control left arrow }
        't': INC(XScreen,10); { control right arrow }
{$IFDEF UseTurboProfessional}
        '': DEC(YScreen,10); { control up arrow; ***TPEnhKbd REQUIRED*** }
        '': INC(YScreen,10); { control down arrow; ***TPEnhKbd REQUIRED*** }
{$ENDIF}
        'w': BEGIN DEC(YScreen,10); DEC(XScreen,10); END; { control home }
        'u': BEGIN INC(YScreen,10); DEC(XScreen,10); END; { control end }
        '': BEGIN DEC(YScreen,10); INC(XScreen,10); END; { control page up }
        'v': BEGIN INC(YScreen,10); INC(XScreen,10); END; { control page down }
      END { END of case statement }
    ELSE flag := TRUE; { end of IF (exkey)... }
  UNTIL flag;

  { alter XScreen and YScreen pixel address for cursor offset; back to
    center of cross hair }
  XScreen := XScreen + Cursor_Xoffset;
  YScreen := YScreen + Cursor_Yoffset;

END; { of Update_Cursor }

{ ****************************************************************** }

PROCEDURE Initialize_CrossHairs;
{
  sets CrossHairs on flag to FALSE which is == off;
}
BEGIN
  { set CrossHairs status flag to off }
  CrossHairs_Onflag := FALSE;
END; { of Initialize_CrossHairs }

{ ****************************************************************** }

PROCEDURE Toggle_CrossHairs(XScreen,YScreen :INTEGER);
{
  this proc toggles CrossHairs status as to whether it is visible; if it
  is on, turn it off; if it is off, turn it on; reset flag accordingly;
  do it at the current position (XScreen,YScreen);
}
BEGIN
  IF (CrossHairs_Onflag) THEN
    CrossHairs_Onflag := FALSE { set flag to CrossHairs off state }
  ELSE
    CrossHairs_Onflag := TRUE; { set flag for CrossHairs on }

  SetWriteMode(XORPut); { set to XOR write mode }
  Line(0,YScreen,XWindMax,YScreen); { draw/erase horizontal hair }
  Line(XScreen,0,XScreen,YWindMax); { draw/erase vertical hair }
  SetWriteMode(CopyPut); { re-enable normal write mode }
END; { of Toggle_CrossHairs }

{ ****************************************************************** }

PROCEDURE Update_CrossHairs(VAR XScreen,YScreen :INTEGER;
                            VAR Keystroke :CHAR);
{
  this routine moves and updates the CrossHairs position; it puts
  updated cross hairs position in (XScreen,YScreen); it assumes on entry
  that the CrossHairs are on and visible; it returns with the ASCII
  character struck in Keystroke, unless the key was a select few of
  the extended scan keys: (see below)
    left/right arrow        - move one screen pixel left/right
    up/down arrow           - move one screen pixel up/down
    home/end                - move one pixel to the left *and* one
                              up/down
    pgup/pgdn               - move one pixel to the right *and* one
                              up/down
    <Ctrl> any of the above - moves the self similar direction but
                              by 10 pixels

  the user should have set the proper linestyle and color in the
  procedure that calls Update_CrossHairs;
  Update_CrossHairs exits with the CrossHairs off (not visible);
}
VAR
  flag,exkey      :BOOLEAN;
BEGIN
  flag:=FALSE; { initialize done flag on entry }

  { if CrossHairs is on, turn it off }
  IF ( CrossHairs_Onflag) THEN Toggle_CrossHairs(XScreen,YScreen);

  REPEAT
    Toggle_CrossHairs(XScreen,YScreen); { draw image }
    ReadKeyStroke(keystroke,exkey); { wait for key to be hit }
    Toggle_CrossHairs(XScreen,YScreen); { erase image }
    IF (exkey) THEN { this an extended key scan code }
      CASE keystroke OF
        'H': DEC(YScreen); { up arrow }
        'P': INC(YScreen); { down arrow }
        'K': DEC(XScreen); { left arrow }
        'M': INC(XScreen); { right arrow }
        'G': BEGIN DEC(YScreen); DEC(XScreen); END; { home key }
        'O': BEGIN INC(YScreen); DEC(XScreen); END; { end key }
        'I': BEGIN DEC(YScreen); INC(XScreen); END; { page up }
        'Q': BEGIN INC(YScreen); INC(XScreen); END; { page down }
        's': DEC(XScreen,10); { control left arrow }
        't': INC(XScreen,10); { control right arrow }
{$IFDEF UseTurboProfessional}
        '': DEC(YScreen,10); { control up arrow; ***TPEnhKbd REQUIRED*** }
        '': INC(YScreen,10); { control down arrow; ***TPEnhKbd REQUIRED*** }
{$ENDIF}
        'w': BEGIN DEC(YScreen,10); DEC(XScreen,10); END; { control home }
        'u': BEGIN INC(YScreen,10); DEC(XScreen,10); END; { control end }
        '': BEGIN DEC(YScreen,10); INC(XScreen,10); END; { control page up }
        'v': BEGIN INC(YScreen,10); INC(XScreen,10); END; { control page down }
      END { END of case statement }
    ELSE flag := TRUE; { end of IF (exkey)... }
  UNTIL flag;

END; { of Update_CrossHairs }

FUNCTION Sign(invalue :FLOAT) :FLOAT;
{
  calculates the Sign of a floating point number
}
BEGIN
  IF (invalue < 0.0) THEN
    Sign := -1.0
  ELSE
    Sign := 1.0;
END; { of Sign }

PROCEDURE Local_Contour(Block_X,X_Step,Block_Y,Y_Step,
                        Num_Contours : INTEGER;
                        Use_Colors :BOOLEAN);
{
  plots all contours within a single cell, using bilinear patch
  approximation.

  Explicit Inputs:
    Block_X       x number of upper left datum for this patch
    X_Step        quantum for X direction grid step
    Block_Y       y number of upper left datum for this patch
    Y_Step        quantum for Y direction grid step
    Num_Contours  number of contour lines specified
    Use_Colors    flag to use colors (=TRUE) or use line_styles (=FALSE)

  Implicit Inputs: (Global Declarations)
    TwoD_Array_Pointer    pointer to 2-D array of data
    Contours              array of contour levels

  Outputs:
    screen using HiGraf routines which use the BGI graphics calls;

  Calls:
    Sign                  Sign of a floating point number

  Optimized by rearranging and recoding a little, but accurate
  profiling has not been done yet. dbh 1/11/89
}
VAR
  x_term,y_term,x_y_term,constant :FLOAT;
  line_number :ARRAY[0..3] OF INTEGER;
  i,number_of_points,contour_number :INTEGER;
  v00,v01,v10,v11 :FLOAT;
  x_array,y_array :ARRAY[0..3] OF FLOAT;
  contour_level,min,max :FLOAT;

PROCEDURE FindXY(Segment_Number :INTEGER;VAR X,Y :FLOAT) ;
{
  calculates the x,y coordinates of a contour line, given which
  segment it penetrates

  Inputs:
    Segment_Number            0=upper segment, 1=left segment,
                              2=right segment, 3=bottom segment
    contour_level             value of contour line
    constant                  constant term of bilinear patch
    x_term                    x linear term of bilinear patch
    y_term                    y linear term of bilinear patch
    x_y_term                  bilinear term of bilinear patch

  Outputs:
    X,Y                       local x,y coordinates of segment
                              intersection
}
BEGIN
  CASE (Segment_Number) OF
        0 : BEGIN { upper segment }
              Y := 0.0;
              X := (contour_level-constant)/x_term;
            END;
        1 : BEGIN { left segment }
              X := 0.0;
              Y := (contour_level-constant)/y_term;
            END;
        2 : BEGIN { right segment }
              X := X_Step;
              y := (contour_level-constant-x_term)/(y_term+x_y_term);
            END;
        3 : BEGIN { bottom segment }
               Y := Y_Step;
               X := (contour_level-constant-y_term)/(x_term+x_y_term);
            END;
  END; { of CASE }
END; { of FindXY }

PROCEDURE Sort_Arrays;
{
  sorts the x,y segment intersections in order of ascending y
  value, via a slow, dumb bubble sort
  inputs:
    x_array,y_array       arrays of segment intersections
  outputs:
    x_array,y_array       sorted arrays of segment intersections
}
VAR
  x_temp,y_temp : FLOAT;
  i,j : integer;
BEGIN
  FOR i:=0 TO 2 DO
    FOR j:=0 TO 2-i DO
      IF (y_array[j]<y_array[j+1]) THEN
      BEGIN
        y_temp := y_array[j];
        x_temp := x_array[j];
        y_array[j] := y_array[j+1];
        x_array[j] := x_array[j+1];
        y_array[j+1] := y_temp;
        x_array[j+1] := x_temp;
      END; { IF }
END; { of Sort_Arrays }

BEGIN { Local_Contour }
  { vxx are used to enhance speed by not requiring pointer arithmetic
   in many places }
  v00 := TwoD_Array_Pointer^[Block_X]^[Block_Y];
  v01 := TwoD_Array_Pointer^[Block_X]^[Block_Y+Y_Step];
  v10 := TwoD_Array_Pointer^[Block_X+X_Step]^[Block_Y];
  v11 := TwoD_Array_Pointer^[Block_X+X_Step]^[Block_Y+Y_Step];

  { set min and max values for this patch to make quick comparisons to
     decide if necessary to draw contour line through this patch }
  min := v00;
  max := v00;
  IF (min>v01) THEN
    min := v01;
  IF (max<v01) THEN
    max := v01;
  IF (min>v10) THEN
    min := v10;
  IF (max<v10) THEN
    max := v10;
  IF (min>v11) THEN
    min := v11;
  IF (max<v11) THEN
    max := v11;

  { calculate the terms of the bilinear equation for this patch }
  constant := v00;
  x_term := v10-constant;
  y_term := v01-constant;
  x_y_term := v11-(x_term+y_term+constant);

  { for each contour line }
  FOR contour_number:= 0 TO Num_Contours-1 DO
  BEGIN
    { in Contour_Plot we have already made sure that the contour
      line is NOT an integer, so that it cannot
      go through a corner of the patch }
    contour_level := Contours[contour_number];
    { if this contour level requires a line in this patch }
    IF ((contour_level > min) AND (contour_level < max))THEN
    BEGIN
      { see how many endpoints there are, either 2 or 4 }
      number_of_points := 0;
      { check top line first }
      IF (Sign(v00-contour_level) <> Sign(v10-contour_level)) THEN
      BEGIN
        line_number[number_of_points] := 0;
        INC(number_of_points);
      END;
      { now check left side line }
      IF (Sign(v00-contour_level) <> Sign(v01-contour_level)) THEN
      BEGIN
        line_number[number_of_points] := 1;
        INC(number_of_points);
      END;
      { now check right side line }
      IF (Sign(v10-contour_level) <> Sign(v11-contour_level)) THEN
      BEGIN
        line_number[number_of_points] := 2;
        INC(number_of_points);
      END;
      { check for bottom is a little easier }
      IF ((number_of_points = 1) OR (number_of_points = 3)) THEN
      BEGIN
        line_number[number_of_points] := 3;
        INC(number_of_points);
      END;
      { if we find a line needs to be drawn }
      IF (number_of_points > 0) THEN
      BEGIN
        IF (Use_Colors) THEN { set the color for this contour number }
          SetColor((contour_number MOD MaxColor)+1)
        ELSE { select the right line_style }
          { we use only 4 user defined line styles }
          SetLineStyle(UserBitLn,Custom_LineStyle[contour_number MOD 4],1);

        FOR  i:=0  TO number_of_points-1  DO
          { then calculate intersection of contour with patch sides }
          FindXY(line_number[i],x_array[i],y_array[i]);
        { if we have only two intersections, just draw the line }
        IF (number_of_points = 2) THEN
        BEGIN
          MoveTo(ROUND((Block_X+x_array[0])*XAxisScale) + ConstX,
                 ROUND((Block_Y+y_array[0])*YAxisScale) + ConstY);
          LineTo(ROUND((Block_X+x_array[1])*XAxisScale) + ConstX,
                 ROUND((Block_Y+y_array[1])*YAxisScale) + ConstY);
        END
        ELSE
        BEGIN
          { if we have 4 intersections (2 lines), then we need to
            sort the intersection points by y to prevent crossing of
            the contours, and to match with bilinear contour }
          Sort_Arrays;
          { then draw the two lines up }
          MoveTo(ROUND((Block_X+x_array[0])*XAxisScale) + ConstX,
                 ROUND((Block_Y+y_array[0])*YAxisScale) + ConstY);
          LineTo(ROUND((Block_X+x_array[1])*XAxisScale) + ConstX,
                 ROUND((Block_Y+y_array[1])*YAxisScale) + ConstY);
          MoveTo(ROUND((Block_X+x_array[2])*XAxisScale) + ConstX,
                 ROUND((Block_Y+y_array[2])*YAxisScale) + ConstY);
          LineTo(ROUND((Block_X+x_array[3])*XAxisScale) + ConstX,
                 ROUND((Block_Y+y_array[3])*YAxisScale) + ConstY);
        END;
      END; { IF number_of_points > 0 }
    END; { IF contour_level > min ... }
  END; { FOR contour_number }
END; { of Local_Contour }

PROCEDURE Contour_Plot(X_Begin, X_End, X_Step,
                       Y_Begin, Y_End, Y_Step,
                       Num_Contours :INTEGER;
                       Use_Colors :BOOLEAN);
{
  plots all contours within data array, using bilinear patch
  approximation.

  Explicit Inputs:
    X_Begin        start position in 2d array
    X_End          ending position in 2d array
    X_Step         quantum for step in X grid direction;
                   must be an even divisor of Max_X_Size!
                   ie for 256, X_Step must = 1,2,4,8, etc
    Y_Begin        start position in 2d array
    Y_End          ending position in 2d array
    Y_Step         quantum for step in Y grid direction
                   must be an even divisor of Max_Y_Size!
                   ie for 256, Y_Step must = 1,2,4,8, etc
    Num_Contours   number of contour lines specified
    Use_Colors     flag to use colors (=TRUE) or use
                   line_styles (=FALSE)

  Implicit Inputs: (Global Declarations)
    TwoD_Array_Pointer    pointer to 2-D array of data
    Contours              array of contour levels ordered in
                          a *specific* way: from lowest
                          (Contours[0]) to highest
                          (Contours[Num_Contours-1])

  Outputs:
    screen (via call to Local_Contour);

  Note:
    calls to HiGraf set up routines (Initialize_Graph & Setup_Graph
    must already have been done before call to Contour_Plot can be
    made

  Calls:
    Local_Contour         plots contour within single patch
}
CONST
  epsilon = 1.0e-3;

VAR
  i,j,stopx,stopy : INTEGER;

BEGIN { Contour_Plot }
  MaxColor := GetMaxColor; { do this only once }
  { adjust contour level if one *happens* to be an integer }
  FOR i:= 0 TO Num_Contours-1 DO
    IF (contours[i] = ROUND(contours[i])) THEN
      contours[i] := contours[i]+epsilon;

  i := X_Begin;
  stopx := X_End-X_Step+1;
  stopy := Y_End-Y_Step+1;
  WHILE (i < stopx) DO
  BEGIN
    j := Y_Begin;
    WHILE (j < stopy) DO
    BEGIN
      Local_Contour(i,X_Step,j,Y_Step,Num_Contours,Use_Colors);
      INC(j,Y_Step);
    END; { WHILE j }
    INC(i,X_Step);
  END; { WHILE i }

  { set pen color and line style back to "normal" after contour plot finishes }
  SetLineStyle(SolidLn,0,NormWidth);
  SetColor(MaxColor);

END; { of Contour_Plot }

PROCEDURE Color_Contour_Plot(X_Begin, X_End, X_Step,
                             Y_Begin, Y_End, Y_Step,
                             Num_Contours :INTEGER);
{
  proc to do "quick and dirty" contour plot using only colors of all
  contours within data array

  Explicit Inputs:
    X_Begin        start position in 2d array
    X_End          ending position in 2d array
    X_Step         not used; only here to look like Contour_Plot call
    Y_Begin        start position in 2d array
    Y_End          ending position in 2d array
    Y_Step         not used; only here to look like Contour_Plot call
    Num_Contours   number of contour lines specified

  Implicit Inputs: (Global Declarations)
    TwoD_Array_Pointer    pointer to 2-D array of data
    Contours              array of contour levels ordered in a
                          *specific* way: from lowest
                          (Contours[1]) to highest
                          (Contours[Num_Contours])

  Outputs:
    screen

  Note:
    calls to HiGraf set up routines (Initialize_Graph & Setup_Graph
    must already have been done before call to Contour_Plot can be
    made

  Calls:
   Bar (from Graph unit) to do actual plotting
       (plus other calls to Graph)

  Optimized by a little recoding; no accurate profiling done yet
  dbh 1/10/89
}
VAR
  i,j,k :INTEGER;
  stopx,stopy :INTEGER;
  x1,y1,x2,y2 :INTEGER;
  x,y :FLOAT;
  temp :LONGINT;
  done :BOOLEAN;
  color :WORD;
BEGIN
  MaxColor := GetMaxColor; { do this only once }
  i := X_Begin;
  stopx := X_End+1;
  stopy := Y_End+1;
  WHILE (i < stopx) DO
  BEGIN
    j := Y_Begin;
    WHILE (j < stopy) DO
    BEGIN
      temp := TwoD_Array_Pointer^[i]^[j]; { get current variable for work }
      done := FALSE;
      color := 0; { set color to black if value is below minimum contour }
      k := 0;
      REPEAT
        IF (Contours[k] > temp) THEN
        BEGIN
          color := (k MOD MaxColor);
          done := TRUE;
        END;
        INC(k);
        IF (k > Num_Contours) THEN done := TRUE;
      UNTIL done;

      { in the interest of speed, check if color <> 0 (black) }
      IF NOT (color = 0) THEN
      BEGIN
        SetFillStyle(SolidFill,color); { set box to solid fill and its color }

        { the next two lines can be commented out in the interest of speed.
          they are necessary for this routine to work exactly right but they
          slow things down by about 10% (!). }
        x := i-0.5*X_Step;
        y := j-0.5*Y_Step;

        x1 := ROUND(x*XAxisScale) + ConstX;
        y1 := ROUND(y*YAxisScale) + ConstY;
        { same as other x and y above; replace x,y in ROUND w/ i+X_Step,j+Y_Step }
        x := x + X_Step;
        y := y + Y_Step;

        x2 := ROUND(x*XAxisScale) + ConstX;
        y2 := ROUND(y*YAxisScale) + ConstY;
        Bar(x1,y1,x2,y2);
      END; { IF NOT (color = 0) }
      INC(j,Y_Step);
    END; { WHILE j }
    INC(i,X_Step);
  END; { WHILE i }

  { set pen color and line style back to "normal" after color plot finishes }
  SetLineStyle(SolidLn,0,NormWidth);
  SetColor(MaxColor);
END; { of Color_Contour_Plot }

{$F+}
PROCEDURE HiGraf_ExitProc;
{
  exit procedure for this unit; all it does is do a CloseGraph to leave
  the system in a sensible screen mode.
}
BEGIN
  ExitProc := HiGraf_ExitSave; { restore link to old exit proc }
  CloseGraph;
END; { of HiGraf_ExitProc }
{$F-}

BEGIN
  TwoD_Array_Pointer := NIL; { init pointer }

  HiGraf_ExitSave := ExitProc; { link in HiGraf exit procedure }
  ExitProc := @HiGraf_ExitProc;

END. { HiGraf }
